@@ -93,7 +93,7 @@ def generate_leaf_updates(leaf_block, dist_thresholds, data, dist):
     return updates
 
 
-@numba.njit(locals={"d": numba.float32, "p": numba.int32, "q": numba.int32}, cache=True)
+@numba.njit(locals={"d": numba.float32, "p": numba.int32, "q": numba.int32}, cache=False)
 def init_rp_tree(data, dist, current_graph, leaf_array):
 
     n_leaves = leaf_array.shape[0]
@@ -137,7 +137,7 @@ def init_rp_tree(data, dist, current_graph, leaf_array):
 @numba.njit(
     fastmath=True,
     locals={"d": numba.float32, "idx": numba.int32, "i": numba.int32},
-    cache=True,
+    cache=False,
 )
 def init_random(n_neighbors, data, heap, dist, rng_state):
     for i in range(data.shape[0]):
@@ -199,7 +199,7 @@ def generate_graph_updates(
     return updates
 
 
-@numba.njit(cache=True)
+@numba.njit(cache=False)
 def process_candidates(
     data,
     dist,
@@ -630,6 +630,13 @@ class NNDescent:
         non-negligible computation cost in building the index. Don't tweak
         this value unless you know what you're doing.
 
+    max_rptree_depth: int (optional, default=100)
+        Maximum depth of random projection trees. Increasing this may result in a
+        richer, deeper random projection forest, but it may be composed of many
+        degenerate branches. Increase leaf_size in order to keep shallower, wider
+        nondegenerate trees. Such wide trees, however, may yield poor performance
+        of the preparation of the NN descent.
+
     n_iters: int (optional, default=None)
         The maximum number of NN-descent iterations to perform. The
         NN-descent algorithm can abort early if limited progress is being
@@ -679,6 +686,7 @@ class NNDescent:
         random_state=None,
         low_memory=True,
         max_candidates=None,
+        max_rptree_depth=200,
         n_iters=None,
         delta=0.001,
         n_jobs=None,
@@ -702,6 +710,7 @@ class NNDescent:
         self.prune_degree_multiplier = pruning_degree_multiplier
         self.diversify_prob = diversify_prob
         self.n_search_trees = n_search_trees
+        self.max_rptree_depth = max_rptree_depth
         self.max_candidates = max_candidates
         self.low_memory = low_memory
         self.n_iters = n_iters
@@ -776,6 +785,7 @@ class NNDescent:
 
         if metric == "dot":
             data = normalize(data, norm="l2", copy=copy_on_normalize)
+            self._raw_data = data
 
         self.rng_state = current_random_state.randint(INT32_MIN, INT32_MAX, 3).astype(
             np.int64
@@ -799,6 +809,7 @@ class NNDescent:
                 current_random_state,
                 self.n_jobs,
                 self._angular_trees,
+                max_depth=self.max_rptree_depth,
             )
             leaf_array = rptree_leaf_array(self._rp_forest)
         else:
@@ -988,6 +999,7 @@ class NNDescent:
                         current_random_state,
                         self.n_jobs,
                         self._angular_trees,
+                        max_depth=self.max_rptree_depth,
                     )
                     self._search_forest = [
                         convert_tree_format(
@@ -1838,6 +1850,7 @@ class NNDescent:
                 current_random_state,
                 self.n_jobs,
                 self._angular_trees,
+                max_depth=self.max_rptree_depth,
             )
             leaf_array = rptree_leaf_array(self._rp_forest)
             current_graph = make_heap(self._raw_data.shape[0], self.n_neighbors)
