@@ -141,6 +141,18 @@ def angular_random_projection_split(data, indices, rng_state):
             side[i] = 1
             n_right += 1
 
+    # If all points end up on one side, something went wrong numerically
+    # In this case, assign points randomly; they are likely very close anyway
+    if n_left == 0 or n_right == 0:
+        n_left = 0
+        n_right = 0
+        for i in range(indices.shape[0]):
+            side[i] = tau_rand_int(rng_state) % 2
+            if side[i] == 0:
+                n_left += 1
+            else:
+                n_right += 1
+
     # Now that we have the counts allocate arrays
     indices_left = np.empty(n_left, dtype=np.int32)
     indices_right = np.empty(n_right, dtype=np.int32)
@@ -248,6 +260,18 @@ def euclidean_random_projection_split(data, indices, rng_state):
             side[i] = 1
             n_right += 1
 
+    # If all points end up on one side, something went wrong numerically
+    # In this case, assign points randomly; they are likely very close anyway
+    if n_left == 0 or n_right == 0:
+        n_left = 0
+        n_right = 0
+        for i in range(indices.shape[0]):
+            side[i] = tau_rand_int(rng_state) % 2
+            if side[i] == 0:
+                n_left += 1
+            else:
+                n_right += 1
+
     # Now that we have the counts allocate arrays
     indices_left = np.empty(n_left, dtype=np.int32)
     indices_right = np.empty(n_right, dtype=np.int32)
@@ -372,6 +396,18 @@ def sparse_angular_random_projection_split(inds, indptr, data, indices, rng_stat
             side[i] = 1
             n_right += 1
 
+    # If all points end up on one side, something went wrong numerically
+    # In this case, assign points randomly; they are likely very close anyway
+    if n_left == 0 or n_right == 0:
+        n_left = 0
+        n_right = 0
+        for i in range(indices.shape[0]):
+            side[i] = tau_rand_int(rng_state) % 2
+            if side[i] == 0:
+                n_left += 1
+            else:
+                n_right += 1
+
     # Now that we have the counts allocate arrays
     indices_left = np.empty(n_left, dtype=np.int32)
     indices_right = np.empty(n_right, dtype=np.int32)
@@ -479,6 +515,18 @@ def sparse_euclidean_random_projection_split(inds, indptr, data, indices, rng_st
             side[i] = 1
             n_right += 1
 
+    # If all points end up on one side, something went wrong numerically
+    # In this case, assign points randomly; they are likely very close anyway
+    if n_left == 0 or n_right == 0:
+        n_left = 0
+        n_right = 0
+        for i in range(indices.shape[0]):
+            side[i] = abs(tau_rand_int(rng_state)) % 2
+            if side[i] == 0:
+                n_left += 1
+            else:
+                n_right += 1
+
     # Now that we have the counts allocate arrays
     indices_left = np.empty(n_left, dtype=np.int32)
     indices_right = np.empty(n_right, dtype=np.int32)
@@ -501,7 +549,6 @@ def sparse_euclidean_random_projection_split(inds, indptr, data, indices, rng_st
 
 @numba.njit(
     nogil=True,
-    cache=True,
     locals={"left_node_num": numba.types.int32, "right_node_num": numba.types.int32},
 )
 def make_euclidean_tree(
@@ -513,8 +560,9 @@ def make_euclidean_tree(
     point_indices,
     rng_state,
     leaf_size=30,
+    max_depth=200,
 ):
-    if indices.shape[0] > leaf_size:
+    if indices.shape[0] > leaf_size and max_depth > 0:
         (
             left_indices,
             right_indices,
@@ -531,6 +579,7 @@ def make_euclidean_tree(
             point_indices,
             rng_state,
             leaf_size,
+            max_depth - 1,
         )
 
         left_node_num = len(point_indices) - 1
@@ -544,6 +593,7 @@ def make_euclidean_tree(
             point_indices,
             rng_state,
             leaf_size,
+            max_depth - 1,
         )
 
         right_node_num = len(point_indices) - 1
@@ -563,7 +613,6 @@ def make_euclidean_tree(
 
 @numba.njit(
     nogil=True,
-    cache=True,
     locals={
         "children": numba.types.ListType(children_type),
         "left_node_num": numba.types.int32,
@@ -579,8 +628,9 @@ def make_angular_tree(
     point_indices,
     rng_state,
     leaf_size=30,
+    max_depth=200,
 ):
-    if indices.shape[0] > leaf_size:
+    if indices.shape[0] > leaf_size and max_depth > 0:
         (
             left_indices,
             right_indices,
@@ -597,6 +647,7 @@ def make_angular_tree(
             point_indices,
             rng_state,
             leaf_size,
+            max_depth - 1,
         )
 
         left_node_num = len(point_indices) - 1
@@ -610,6 +661,7 @@ def make_angular_tree(
             point_indices,
             rng_state,
             leaf_size,
+            max_depth - 1,
         )
 
         right_node_num = len(point_indices) - 1
@@ -629,7 +681,6 @@ def make_angular_tree(
 
 @numba.njit(
     nogil=True,
-    cache=True,
     locals={"left_node_num": numba.types.int32, "right_node_num": numba.types.int32},
 )
 def make_sparse_euclidean_tree(
@@ -643,8 +694,9 @@ def make_sparse_euclidean_tree(
     point_indices,
     rng_state,
     leaf_size=30,
+    max_depth=200,
 ):
-    if indices.shape[0] > leaf_size:
+    if indices.shape[0] > leaf_size and max_depth > 0:
         (
             left_indices,
             right_indices,
@@ -665,6 +717,7 @@ def make_sparse_euclidean_tree(
             point_indices,
             rng_state,
             leaf_size,
+            max_depth - 1,
         )
 
         left_node_num = len(point_indices) - 1
@@ -680,6 +733,7 @@ def make_sparse_euclidean_tree(
             point_indices,
             rng_state,
             leaf_size,
+            max_depth - 1,
         )
 
         right_node_num = len(point_indices) - 1
@@ -699,7 +753,6 @@ def make_sparse_euclidean_tree(
 
 @numba.njit(
     nogil=True,
-    cache=True,
     locals={"left_node_num": numba.types.int32, "right_node_num": numba.types.int32},
 )
 def make_sparse_angular_tree(
@@ -713,8 +766,9 @@ def make_sparse_angular_tree(
     point_indices,
     rng_state,
     leaf_size=30,
+    max_depth=200,
 ):
-    if indices.shape[0] > leaf_size:
+    if indices.shape[0] > leaf_size and max_depth > 0:
         (
             left_indices,
             right_indices,
@@ -735,6 +789,7 @@ def make_sparse_angular_tree(
             point_indices,
             rng_state,
             leaf_size,
+            max_depth - 1,
         )
 
         left_node_num = len(point_indices) - 1
@@ -750,6 +805,7 @@ def make_sparse_angular_tree(
             point_indices,
             rng_state,
             leaf_size,
+            max_depth - 1,
         )
 
         right_node_num = len(point_indices) - 1
@@ -765,8 +821,8 @@ def make_sparse_angular_tree(
         point_indices.append(indices)
 
 
-@numba.njit(nogil=True, cache=True)
-def make_dense_tree(data, rng_state, leaf_size=30, angular=False):
+@numba.njit(nogil=True)
+def make_dense_tree(data, rng_state, leaf_size=30, angular=False, max_depth=200):
     indices = np.arange(data.shape[0]).astype(np.int32)
 
     hyperplanes = numba.typed.List.empty_list(dense_hyperplane_type)
@@ -784,6 +840,7 @@ def make_dense_tree(data, rng_state, leaf_size=30, angular=False):
             point_indices,
             rng_state,
             leaf_size,
+            max_depth=max_depth,
         )
     else:
         make_euclidean_tree(
@@ -795,14 +852,28 @@ def make_dense_tree(data, rng_state, leaf_size=30, angular=False):
             point_indices,
             rng_state,
             leaf_size,
+            max_depth=max_depth,
         )
 
-    result = FlatTree(hyperplanes, offsets, children, point_indices, leaf_size)
+    max_leaf_size = leaf_size
+    for points in point_indices:
+        if len(points) > max_leaf_size:
+            max_leaf_size = numba.int32(len(points))
+
+    result = FlatTree(hyperplanes, offsets, children, point_indices, max_leaf_size)
     return result
 
 
-@numba.njit(nogil=True, cache=True)
-def make_sparse_tree(inds, indptr, spdata, rng_state, leaf_size=30, angular=False):
+@numba.njit(nogil=True)
+def make_sparse_tree(
+    inds,
+    indptr,
+    spdata,
+    rng_state,
+    leaf_size=30,
+    angular=False,
+    max_depth=200,
+):
     indices = np.arange(indptr.shape[0] - 1).astype(np.int32)
 
     hyperplanes = numba.typed.List.empty_list(sparse_hyperplane_type)
@@ -822,6 +893,7 @@ def make_sparse_tree(inds, indptr, spdata, rng_state, leaf_size=30, angular=Fals
             point_indices,
             rng_state,
             leaf_size,
+            max_depth=max_depth,
         )
     else:
         make_sparse_euclidean_tree(
@@ -835,9 +907,15 @@ def make_sparse_tree(inds, indptr, spdata, rng_state, leaf_size=30, angular=Fals
             point_indices,
             rng_state,
             leaf_size,
+            max_depth=max_depth,
         )
 
-    return FlatTree(hyperplanes, offsets, children, point_indices, leaf_size)
+    max_leaf_size = leaf_size
+    for points in point_indices:
+        if len(points) > max_leaf_size:
+            max_leaf_size = numba.int32(len(points))
+
+    return FlatTree(hyperplanes, offsets, children, point_indices, max_leaf_size)
 
 
 @numba.njit(
@@ -956,6 +1034,7 @@ def make_forest(
     random_state,
     n_jobs=None,
     angular=False,
+    max_depth=200,
 ):
     """Build a random projection forest with ``n_trees``.
 
@@ -993,12 +1072,19 @@ def make_forest(
                     rng_states[i],
                     leaf_size,
                     angular,
+                    max_depth=max_depth,
                 )
                 for i in range(n_trees)
             )
         else:
             result = joblib.Parallel(n_jobs=n_jobs, require="sharedmem")(
-                joblib.delayed(make_dense_tree)(data, rng_states[i], leaf_size, angular)
+                joblib.delayed(make_dense_tree)(
+                    data,
+                    rng_states[i],
+                    leaf_size,
+                    angular,
+                    max_depth=max_depth
+                )
                 for i in range(n_trees)
             )
     except (RuntimeError, RecursionError, SystemError):
@@ -1011,14 +1097,14 @@ def make_forest(
     return tuple(result)
 
 
-@numba.njit(nogil=True, cache=True)
-def get_leaves_from_tree(tree):
+@numba.njit(nogil=True)
+def get_leaves_from_tree(tree, max_leaf_size):
     n_leaves = 0
     for i in range(len(tree.children)):
         if tree.children[i][0] == -1 and tree.children[i][1] == -1:
             n_leaves += 1
 
-    result = np.full((n_leaves, tree.leaf_size), -1, dtype=np.int32)
+    result = np.full((n_leaves, max_leaf_size), -1, dtype=np.int32)
     leaf_index = 0
     for i in range(len(tree.indices)):
         if tree.children[i][0] == -1 or tree.children[i][1] == -1:
@@ -1030,8 +1116,9 @@ def get_leaves_from_tree(tree):
 
 
 def rptree_leaf_array_parallel(rp_forest):
+    max_leaf_size = np.max([rp_tree.leaf_size for rp_tree in rp_forest])
     result = joblib.Parallel(n_jobs=-1, require="sharedmem")(
-        joblib.delayed(get_leaves_from_tree)(rp_tree) for rp_tree in rp_forest
+        joblib.delayed(get_leaves_from_tree)(rp_tree, max_leaf_size) for rp_tree in rp_forest
     )
     return result
 
@@ -1047,7 +1134,6 @@ def rptree_leaf_array(rp_forest):
 def recursive_convert(
     tree, hyperplanes, offsets, children, indices, node_num, leaf_start, tree_node
 ):
-
     if tree.children[tree_node][0] < 0:
         leaf_end = leaf_start + len(tree.indices[tree_node])
         children[node_num, 0] = -leaf_start
@@ -1087,7 +1173,6 @@ def recursive_convert(
 def recursive_convert_sparse(
     tree, hyperplanes, offsets, children, indices, node_num, leaf_start, tree_node
 ):
-
     if tree.children[tree_node][0] < 0:
         leaf_end = leaf_start + len(tree.indices[tree_node])
         children[node_num, 0] = -leaf_start
@@ -1176,7 +1261,6 @@ FLAT_TREE_LEAF_SIZE = 4
 
 
 def denumbaify_tree(tree):
-
     result = (
         tree.hyperplanes,
         tree.offsets,
@@ -1189,7 +1273,6 @@ def denumbaify_tree(tree):
 
 
 def renumbaify_tree(tree):
-
     result = FlatTree(
         tree[FLAT_TREE_HYPERPLANES],
         tree[FLAT_TREE_OFFSETS],
